import collections.abc as _abc
import dataclasses as _dc
import enum as _enum
import pathlib as _pl
import typing as _tp

import pandas as _pd
import numpy as _np
import matplotlib as _mpl
import matplotlib.pyplot as _plt

from optihood.entities import NodeKeys as sheets, BusTypes
from optihood.entities import TransformerLabels as trafo
from optihood.entities import StorageLabels as store
from optihood.entities import SolarLabels as solar
from optihood.entities import SolarTypes
from optihood.entities import LinksLabels as link
from optihood.entities import StorageTypes


class ScenarioDataTypes(_enum.StrEnum):
    example: str = 'example'


class EnergyTypes(_enum.StrEnum):
    electricity = 'electricity'
    domestic_hot_water = "DHW"
    space_heating = "SH"
    gas = 'gas'
    oil = 'oil'
    hydrogen = 'H2'
    biomass = 'biomass'
    low_temp_grid = 'low_temp_grid'
    sh_temp_grid = 'sh_temp_grid'
    med_temp_grid = 'med_temp_grid'
    dhw_temp_grid = 'dhw_temp_grid'
    district = 'district'
    cooling = 'cooling'
    unknown = 'unknown'


def get_energy_type(label: str):
    low_temp_grid_parts = ["lowTemp", "lowTSourceBus", "lowTSinkBus", "lakePotential", "riverPotential"]
    if any(flag in label for flag in low_temp_grid_parts):
        return EnergyTypes.low_temp_grid

    # In the optimization mode with multiple temperature levels, there are three discrete levels marked as:
    # 0 (lowest temperature), 1 and 2 (highest temperature).
    # The lowest temperature (level 0) is the temperature at which space heat is provided
    # The highest temperature (level 2) is the temperature at which domestic hot water is provided
    # The temperature level 1 represents a discrete temperature level between level 0 and level 1.

    sh_temp_grid_parts = ["heat0", "heatLink0", "heatStorageBus0", "heatBus0", "heatDemand0", "heatDemandBus0", "heatDemandEndBus0"]
    if any(flag in label for flag in sh_temp_grid_parts):
        return EnergyTypes.sh_temp_grid

    med_temp_grid_parts = ["heatStorageBus1"]
    if any(flag in label for flag in med_temp_grid_parts):
        return EnergyTypes.med_temp_grid

    dhw_temp_grid_parts = ["heat2", "heatLink2", "heatStorageBus2", "heatBus2", "heatDemand2", "heatDemandBus2", "heatDemandEndBus2"]
    if any(flag in label for flag in dhw_temp_grid_parts):
        return EnergyTypes.dhw_temp_grid

    electric_parts = ["electric", "Electric", "grid", "pv"]
    if any(flag in label for flag in electric_parts):
        return EnergyTypes.electricity

    sh_parts = ["sh", "spaceHeating", "chiller", 'heatSource35']
    # TODO: check why "chiller" is added to sh_parts
    if any(flag in label for flag in sh_parts):
        return EnergyTypes.space_heating

    dhw_parts = ["dhw", "domesticHotWater", 'heatSource65']
    if any(flag in label for flag in dhw_parts):
        return EnergyTypes.domestic_hot_water

    district_parts = ["district", "District", "dh", "heatingSource"]
    if any(flag in label for flag in district_parts):
        return EnergyTypes.district

    gas_parts = ["gas", "Gas"]
    if any(flag in label for flag in gas_parts):
        return EnergyTypes.gas

    biomass_parts = ["biomass", "Biomass"]
    if any(flag in label for flag in biomass_parts):
        return EnergyTypes.biomass

    oil_parts = ["oil", "Oil"]
    if any(flag in label for flag in oil_parts):
        return EnergyTypes.oil

    cooling_parts = ["cool"]
    if any(flag in label for flag in cooling_parts):
        return EnergyTypes.cooling

    print(f'Unknown label type found: {label}')
    return EnergyTypes.unknown


def get_energy_type_based_on_both_labels(label: str, other_label: str) -> EnergyTypes:
    """ Edges are based on the current node and another (to or from)."""

    transformer_parts = ["HP", "solar", "GasBoiler", "BiomassBoiler", "OilBoiler", "Dummy", "HP_d", "GWHP_d", "Chiller"]
    source_parts = ["Dummy", "wasteheatPotential", "wasteIncineratorPotential"]
    storage_parts = [s.value for s in StorageTypes if s not in [StorageTypes.shStorage, StorageTypes.dhwStorage, StorageTypes.electricalStorage]]

    all_parts = transformer_parts + source_parts + storage_parts

    if any(flag in label for flag in all_parts):
        return get_energy_type(other_label)

    return get_energy_type(label)


class MplColorHelper:
    """ Taken from https://stackoverflow.com/questions/26108436/how-can-i-get-the-matplotlib-rgb-color-given-the-colormap-name-boundrynorm-an#26109298
    """

    def __init__(self, cmap_name: str, start_val: _tp.Union[int, float], stop_val: _tp.Union[int, float]):
        self.cmap_name = cmap_name
        self.stop_val = stop_val
        self.cmap = _plt.get_cmap(cmap_name)
        self.norm = _mpl.colors.Normalize(vmin=start_val, vmax=stop_val)
        self.scalarMap = _mpl.cm.ScalarMappable(norm=self.norm, cmap=self.cmap)

    def get_rgb(self, val: _tp.Union[int, float, _np.ndarray]) -> _np.ndarray:
        val = val % self.stop_val
        return self.scalarMap.to_rgba(val)

    def get_hex(self, val: _tp.Union[int, float, _np.ndarray]) -> str:
        # val = val % self.stop_val
        return _mpl.colors.to_hex(self.get_rgb(val))


@_dc.dataclass()
class ScenarioToVisualizerAbstract:
    """ From node and to node could also use enums. """
    label: str
    from_node: _tp.Optional[_tp.Union[str, _abc.Sequence[str]]]  # sources do not have this
    to_node: _tp.Optional[_tp.Union[str, _abc.Sequence[str]]]  # sinks do not have this
    active: bool
    edges_into_node: list[dict[str, dict[str, _tp.Union[str, float, int]]]] = _dc.field(init=False)
    edges_out_of_node: list[dict[str, dict[str, _tp.Union[str, float, int]]]] = _dc.field(init=False)
    color: str = _dc.field(init=False)

    def __post_init__(self):
        colorHelper = MplColorHelper('Paired', 0, 12)
        if hasattr(self, 'building'):
            self.color = colorHelper.get_hex(self.building)
        else:
            self.color = colorHelper.get_hex(0)

    @property
    def id(self):
        """ The id does not show up in the image.
        It is used to identify those nodes, to which the edges connect. """

        if not hasattr(self, "building"):
            return self.label

        return self.get_id_with_building(self.label, self.building)

    @staticmethod
    def get_id_with_building(label: str, building: str) -> str:
        flags_to_be_ignored = ['link', 'Link']

        if any(flag in label for flag in flags_to_be_ignored):
            return label

        zero_padding_level = 3
        return f"{label}_B{str(building).zfill(zero_padding_level)}"

    def get_nodal_infos(self) -> _tp.Optional[dict[str, dict[str, _tp.Union[str, int, float, _pl.Path]]]]:
        raise NotImplementedError('Do not access parent class directly')

    def get_edge_infos(self) -> list[dict[str, dict[str, _tp.Union[str, float, int]]]]:
        if not self.active:
            return []

        self.edges_into_node = []
        if self.from_node:
            if not isinstance(self.from_node, list):
                self.from_node = [self.from_node]
            for from_node in self.from_node:
                energy_type = get_energy_type_based_on_both_labels(self.id, from_node)
                self.edges_into_node.append({'data': {'source': from_node, 'target': self.id},
                                             "classes": energy_type.value})
        self.edges_out_of_node = []
        if self.to_node:
            if not isinstance(self.to_node, list):
                self.to_node = [self.to_node]
            for to_node in self.to_node:
                energy_type = get_energy_type_based_on_both_labels(self.id, to_node)
                self.edges_out_of_node.append({'data': {'source': self.id, 'target': to_node},
                                               "classes": energy_type.value})

        all_edges = self.edges_into_node + self.edges_out_of_node
        return all_edges

    @staticmethod
    def read_nodal_infos(data: dict[str, _tp.Union[str, float, int]]) -> _tp.Optional[str]:
        """ Adding line breaks for every entry would look cleaner. """
        return f"{data}"

    @staticmethod
    def read_edge_infos(data: dict[str, _tp.Union[str, float, int]]):
        """ This may never be needed. """
        raise NotImplementedError('Do not access parent class directly')

    @staticmethod
    def set_from_dataFrame(df: _pd.DataFrame, nr_of_buildings: int):  # -> _tp.Type[ScenarioToVisualizerAbstract]
        """ Typing does not allow usage of this class's type."""
        raise NotImplementedError('Do not access parent class directly')

    @staticmethod
    def get_to_and_from_nodes(line):
        building = line['building']
        to_nodes = line['to'].split(sep=',')
        from_nodes = line['from'].split(sep=',')

        for iNode, from_node in enumerate(from_nodes):
            from_nodes[iNode] = ScenarioToVisualizerAbstract.get_id_with_building(from_node, building)

        for iNode, to_node in enumerate(to_nodes):
            to_nodes[iNode] = ScenarioToVisualizerAbstract.get_id_with_building(to_node, building)

        return from_nodes, to_nodes

    @staticmethod
    def replace_nans(df):
        df.replace({_np.nan: None}, inplace=True)


def scenario_data_factory(scenario_data_type: str) -> _tp.Optional[_tp.Type[ScenarioToVisualizerAbstract]]:
    scenario_data_types = {ScenarioDataTypes.example: NodalDataExample,
                           sheets.buses: BusesConverter,
                           sheets.commodity_sources: CommoditySourcesConverter,
                           sheets.demand: DemandConverter,
                           sheets.grid_connection: GridConnectionConverter,
                           sheets.transformers: TransformersConverter,
                           sheets.storages: StoragesConverter,
                           sheets.solar: SolarConverter,
                           sheets.links: LinksConverter}

    if scenario_data_type not in scenario_data_types:
        # raise NotImplementedError("received unexpected type")
        return  # <- until all sheets are implemented

    return scenario_data_types[scenario_data_type]


@_dc.dataclass()
class NodalDataExample(ScenarioToVisualizerAbstract):
    longitude: float
    latitude: float
    _id: str = _dc.field(init=False)

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    def get_nodal_infos(self):
        return {
            'data': {'id': self.id, 'label': self.label, "lat": self.latitude, "long": self.longitude},
            'position': {'x': 20 * self.latitude, 'y': -20 * self.longitude}
        }


@_dc.dataclass()
class CommoditySourcesConverter(ScenarioToVisualizerAbstract):
    building: int
    variable_costs: _tp.Union[float, _pl.Path]
    CO2_impact: _tp.Union[float, _pl.Path]

    def __post_init__(self):
        super().__post_init__()
        if self.from_node:
            raise Warning(f'Commodity sources do not have a from node. Received {self.from_node}.')

    def get_nodal_infos(self) -> _tp.Optional[dict[str, dict[str, _tp.Union[str, int, float, _pl.Path]]]]:
        if self.active:
            return {"data": {'id': self.id, 'label': self.label, "building": self.building,
                             "variable_costs": self.variable_costs, "CO2_impact": self.CO2_impact, 'color': self.color},
                    'classes': 'source'}

    @staticmethod
    def set_from_dataFrame(df: _pd.DataFrame, nr_of_buildings: int) -> _abc.Sequence[_tp.Type[ScenarioToVisualizerAbstract]]:
        list_of_demands = []

        ScenarioToVisualizerAbstract.replace_nans(df)

        if 'active' not in df.columns:
            df['active'] = True

        for i, line in df.iterrows():
            if not line['active']:
                continue

            to_nodes = line['to'].split(sep=',')
            building = line['building']

            for iNode, to_node in enumerate(to_nodes):
                to_nodes[iNode] = ScenarioToVisualizerAbstract.get_id_with_building(to_node, building)

            list_of_demands.append(CommoditySourcesConverter(line['label'], None,
                                                             to_nodes,
                                                             active=line['active'], building=building,
                                                             variable_costs=line['variable costs'],
                                                             CO2_impact=line['CO2 impact']))
        return list_of_demands


@_dc.dataclass()
class BusesConverter(ScenarioToVisualizerAbstract):
    building: int
    excess: int
    excess_costs: _tp.Union[float, _pl.Path]
    shortage: _tp.Optional[bool] = None
    shortage_costs: _tp.Optional[_tp.Union[float, _pl.Path]] = None

    def __post_init__(self):
        super().__post_init__()
        if self.from_node:
            raise Warning(f'Buses tend not to have a from node assigned in the scenario. Received {self.from_node}.')
        if self.to_node:
            raise Warning(f'Buses tend not to have a from node assigned in the scenario. Received {self.from_node}.')

    def get_nodal_infos(self) -> _tp.Optional[dict[str, dict[str, _tp.Union[str, int, float, _pl.Path]]]]:
        if self.active:
            return {"data": {'id': self.id, 'label': self.label, "building": self.building,
                             "excess": self.excess, "excess_costs": self.excess_costs, "shortage": self.shortage,
                             "shortage_costs": self.shortage_costs, 'color': self.color},
                    'classes': 'bus'}

    def get_edge_infos(self) -> list[dict[str, dict[str, _tp.Union[str, float, int]]]]:
        # TODO: fix building node as dummy and chiller not connecting properly.
        if not self.active:
            return []

        all_normal_edges = super().get_edge_infos()

        excess_edges = []
        if self.excess:
            excess_node_label = 'excess_' + self.id
            energy_type = get_energy_type_based_on_both_labels(self.id, excess_node_label)
            excess_edges.append({'data': {'source': self.id, 'target': excess_node_label},
                                 "classes": energy_type.value})

        shortage_edges = []
        if self.shortage:
            shortage_node_label = 'shortage_' + self.id
            energy_type = get_energy_type_based_on_both_labels(shortage_node_label, self.id)
            excess_edges.append({'data': {'source': shortage_node_label, 'target': self.id},
                                 "classes": energy_type.value})

        self.edges_out_of_node += excess_edges + shortage_edges

        return all_normal_edges + excess_edges + shortage_edges

    @staticmethod
    def set_from_dataFrame(df: _pd.DataFrame, nr_of_buildings: int) -> _abc.Sequence[_tp.Type[ScenarioToVisualizerAbstract]]:
        list_of_buses = []

        ScenarioToVisualizerAbstract.replace_nans(df)

        # TODO: check whether shortage is given without shortage costs.
        if 'shortage' not in df.columns:
            df['shortage'] = None

        if 'shortage costs' not in df.columns:
            df['shortage costs'] = None

        if 'excess' not in df.columns:
            df['excess'] = None

        if 'excess costs' not in df.columns:
            df['excess costs'] = None

        if 'active' not in df.columns:
            df['active'] = True

        for i, line in df.iterrows():
            if not line['active']:
                continue

            list_of_buses.append(
                BusesConverter(line['label'], None, None,
                               active=line['active'], building=line['building'],
                               excess=line['excess'],
                               excess_costs=line['excess costs'],
                               shortage=line['shortage'],
                               shortage_costs=line['shortage costs']
                               ))

            if line['excess']:
                list_of_buses.append(
                    DemandConverter("excess_" + line['label'], None, None,
                                    active=line['active'], building=line['building'],
                                    fixed=line['excess costs'],
                                    nominal_value=0)

                )

            if line['shortage']:
                list_of_buses.append(
                    CommoditySourcesConverter("shortage_" + line['label'], None, None,
                                              active=line['active'], building=line['building'],
                                              variable_costs=line['shortage costs'],
                                              CO2_impact=0)

                )

        return list_of_buses


@_dc.dataclass()
class DemandConverter(ScenarioToVisualizerAbstract):
    building: int
    fixed: int
    nominal_value: int
    building_model: _tp.Optional[bool] = None
    building_model_out: _tp.Optional[str] = None

    def __post_init__(self):
        super().__post_init__()
        if self.to_node:
            raise Warning(f'Buses tend not to have a from node assigned in the scenario. Received {self.from_node}.')

    def get_nodal_infos(self) -> _tp.Optional[dict[str, dict[str, _tp.Union[str, int, float, _pl.Path]]]]:
        if self.active:
            return {"data": {'id': self.id, 'label': self.label, "building": self.building,
                             "fixed": self.fixed, "nominal_value": self.nominal_value, 'color': self.color,
                             "building model": self.building_model, "building model out": self.building_model_out},
                    "classes": "demand"}

    def get_edge_infos(self) -> list[dict[str, dict[str, _tp.Union[str, float, int]]]]:
        if not self.active:
            return []

        all_normal_edges = super().get_edge_infos()

        building_model_edges = []
        if self.building_model_out:
            if not isinstance(self.building_model_out, list):
                self.building_model_out = [self.building_model_out]
            for building_model_out in self.building_model_out:
                energy_type = get_energy_type_based_on_both_labels(self.id, building_model_out)
                building_model_edges.append({'data': {'source': self.id, 'target': building_model_out},
                                             "classes": energy_type.value})

        self.edges_out_of_node += building_model_edges

        return all_normal_edges + building_model_edges

    @staticmethod
    def set_from_dataFrame(df: _pd.DataFrame, nr_of_buildings: int) -> _abc.Sequence[_tp.Type[ScenarioToVisualizerAbstract]]:
        list_of_demands = []

        ScenarioToVisualizerAbstract.replace_nans(df)

        if 'active' not in df.columns:
            df['active'] = True

        if 'building model' not in df.columns:
            df['building model'] = None

        if 'building model out' not in df.columns:
            df['building model out'] = None

        for i, line in df.iterrows():
            if not line['active']:
                continue

            if not line['building model'] == 'yes' and not line['building model'] == 'Yes':
                line['building model out'] = None
                building_model_out_nodes = None

            building = line['building']
            from_nodes = line['from'].split(sep=',')

            for iNode, from_node in enumerate(from_nodes):
                from_nodes[iNode] = ScenarioToVisualizerAbstract.get_id_with_building(from_node, building)

            if line['building model out']:
                building_model_out_nodes = line['building model out'].split(sep=',')

                for iNode, building_model_out in enumerate(building_model_out_nodes):
                    building_model_out_nodes[iNode] = ScenarioToVisualizerAbstract.get_id_with_building(
                        building_model_out, building)

            list_of_demands.append(DemandConverter(line['label'], from_nodes,
                                                   None,
                                                   active=line['active'], building=line['building'],
                                                   fixed=line['fixed'], nominal_value=line['nominal value'],
                                                   building_model=line['building model'],
                                                   building_model_out=building_model_out_nodes))
        return list_of_demands


@_dc.dataclass()
class GridConnectionConverter(ScenarioToVisualizerAbstract):
    building: int
    efficiency: float

    def __post_init__(self):
        super().__post_init__()

    def get_nodal_infos(self) -> _tp.Optional[dict[str, dict[str, _tp.Union[str, int, float, _pl.Path]]]]:
        if self.active:
            return {"data": {'id': self.id, 'label': self.label, "building": self.building,
                             "efficiency": self.efficiency, 'color': self.color},
                    "classes": "grid_connection"}

    @staticmethod
    def set_from_dataFrame(df: _pd.DataFrame, nr_of_buildings: int) -> _abc.Sequence[_tp.Type[ScenarioToVisualizerAbstract]]:
        list_of_demands = []

        ScenarioToVisualizerAbstract.replace_nans(df)

        if 'active' not in df.columns:
            df['active'] = True

        for i, line in df.iterrows():
            if not line['active']:
                continue

            if not GridConnectionConverter.has_building_model(line):
                line['building model out'] = None

            building = line['building']
            to_nodes = line['to'].split(sep=',')
            from_nodes = line['from'].split(sep=',')

            for iNode, from_node in enumerate(from_nodes):
                from_nodes[iNode] = ScenarioToVisualizerAbstract.get_id_with_building(from_node, building)

            for iNode, to_node in enumerate(to_nodes):
                to_nodes[iNode] = ScenarioToVisualizerAbstract.get_id_with_building(to_node, building)

            list_of_demands.append(GridConnectionConverter(line['label'], from_nodes,
                                                           to_nodes,
                                                           active=line['active'],
                                                           building=line['building'],
                                                           efficiency=line['efficiency']))
        return list_of_demands

    @staticmethod
    def has_building_model(line):
        if 'building model' not in line.keys():
            return False

        if not line['building model'] == 'yes' and not line['building model'] == 'Yes':
            return False

        return True


@_dc.dataclass()
class TransformersConverter(ScenarioToVisualizerAbstract):
    building: int
    efficiency: float
    capacity_DHW: float
    capacity_SH: float
    capacity_min: float
    lifetime: float
    maintenance: float
    installation: float
    planification: float
    invest_base: float
    invest_cap: float
    heat_impact: float
    elec_impact: float
    impact_cap: float
    capacity_el: _tp.Optional[float] = None

    def __post_init__(self):
        super().__post_init__()

    def get_nodal_infos(self) -> _tp.Optional[dict[str, dict[str, _tp.Union[str, int, float, _pl.Path]]]]:
        if self.active:
            return {"data": {'id': self.id, 'label': self.label, trafo.building.value: self.building,
                             trafo.efficiency.value: self.efficiency,
                             trafo.capacity_DHW.value: self.capacity_DHW,
                             trafo.capacity_SH.value: self.capacity_SH,
                             trafo.capacity_min.value: self.capacity_min,
                             trafo.lifetime.value: self.lifetime,
                             trafo.maintenance.value: self.maintenance,
                             trafo.installation.value: self.installation,
                             trafo.planification.value: self.planification,
                             trafo.invest_base.value: self.invest_base,
                             trafo.invest_cap.value: self.invest_cap,
                             trafo.heat_impact.value: self.heat_impact,
                             trafo.elec_impact.value: self.elec_impact,
                             trafo.impact_cap.value: self.impact_cap,
                             trafo.capacity_el.value: self.capacity_el,
                             'color': self.color,
                             },
                    "classes": "transformer"}

    @staticmethod
    def set_from_dataFrame(df: _pd.DataFrame, nr_of_buildings: int) -> _abc.Sequence[_tp.Type[ScenarioToVisualizerAbstract]]:
        list_of_demands = []

        ScenarioToVisualizerAbstract.replace_nans(df)

        if 'active' not in df.columns:
            df['active'] = True

        if 'capacity_el' not in df.columns:
            df['capacity_el'] = None

        for i, line in df.iterrows():
            if not line['active']:
                continue

            building = line['building']
            to_nodes = line['to'].split(sep=',')
            from_nodes = line['from'].split(sep=',')

            for iNode, from_node in enumerate(from_nodes):
                from_nodes[iNode] = ScenarioToVisualizerAbstract.get_id_with_building(from_node, building)

            for iNode, to_node in enumerate(to_nodes):
                to_nodes[iNode] = ScenarioToVisualizerAbstract.get_id_with_building(to_node, building)

            list_of_demands.append(
                TransformersConverter(line['label'], from_nodes,
                                      to_nodes, active=line['active'],
                                      building=line[trafo.building.value],
                                      efficiency=line[trafo.efficiency.value],
                                      capacity_DHW=line[trafo.capacity_DHW.value],
                                      capacity_SH=line[trafo.capacity_SH.value],
                                      capacity_min=line[trafo.capacity_min.value],
                                      lifetime=line[trafo.lifetime.value],
                                      maintenance=line[trafo.maintenance.value],
                                      installation=line[trafo.installation.value],
                                      planification=line[trafo.planification.value],
                                      invest_base=line[trafo.invest_base.value],
                                      invest_cap=line[trafo.invest_cap.value],
                                      heat_impact=line[trafo.heat_impact.value],
                                      elec_impact=line[trafo.elec_impact.value],
                                      impact_cap=line[trafo.impact_cap.value]
                                      ))
        return list_of_demands


@_dc.dataclass()
class StoragesConverter(ScenarioToVisualizerAbstract):
    building: int
    efficiency_inflow: _tp.Optional[float]
    efficiency_outflow: _tp.Optional[float]
    initial_capacity: float
    capacity_min: float
    capacity_max: float
    capacity_loss: _tp.Optional[float]
    lifetime: float
    maintenance: float
    installation: float
    planification: float
    invest_base: float
    invest_cap: float
    heat_impact: float
    elec_impact: float
    impact_cap: float

    def __post_init__(self):
        super().__post_init__()

    def get_nodal_infos(self) -> _tp.Optional[dict[str, dict[str, _tp.Union[str, int, float, _pl.Path]]]]:
        if self.active:
            return {"data": {'id': self.id, 'label': self.label, store.building.value: self.building,
                             store.efficiency_inflow.value: self.efficiency_inflow,
                             store.efficiency_outflow.value: self.efficiency_outflow,
                             store.initial_capacity.value: self.initial_capacity,
                             store.capacity_min.value: self.capacity_min,
                             store.capacity_max.value: self.capacity_max,
                             store.capacity_loss.value: self.capacity_loss,
                             store.lifetime.value: self.lifetime,
                             store.maintenance.value: self.maintenance,
                             store.installation.value: self.installation,
                             store.planification.value: self.planification,
                             store.invest_base.value: self.invest_base,
                             store.invest_cap.value: self.invest_cap,
                             store.heat_impact.value: self.heat_impact,
                             store.elec_impact.value: self.elec_impact,
                             store.impact_cap.value: self.impact_cap,
                             'color': self.color,
                             },
                    "classes": "storage"}

    @staticmethod
    def set_from_dataFrame(df: _pd.DataFrame, nr_of_buildings: int) -> _abc.Sequence[_tp.Type[ScenarioToVisualizerAbstract]]:
        list_of_demands = []

        ScenarioToVisualizerAbstract.replace_nans(df)

        if 'active' not in df.columns:
            df['active'] = True

        if store.efficiency_inflow not in df.columns:
            df[store.efficiency_inflow] = None

        if store.efficiency_outflow not in df.columns:
            df[store.efficiency_outflow] = None

        if store.capacity_loss not in df.columns:
            df[store.capacity_loss] = None

        for i, line in df.iterrows():
            if not line['active']:
                continue

            building = line['building']
            to_nodes = line['to'].split(sep=',')
            from_nodes = line['from'].split(sep=',')

            for iNode, from_node in enumerate(from_nodes):
                from_nodes[iNode] = ScenarioToVisualizerAbstract.get_id_with_building(from_node, building)

            for iNode, to_node in enumerate(to_nodes):
                to_nodes[iNode] = ScenarioToVisualizerAbstract.get_id_with_building(to_node, building)

            list_of_demands.append(
                StoragesConverter(line['label'], from_nodes,
                                  to_nodes,
                                  building=line[store.building.value], active=True,
                                  efficiency_inflow=line[store.efficiency_inflow.value],
                                  efficiency_outflow=line[store.efficiency_outflow.value],
                                  initial_capacity=line[store.initial_capacity.value],
                                  capacity_min=line[store.capacity_min.value],
                                  capacity_max=line[store.capacity_max.value],
                                  capacity_loss=line[store.capacity_loss.value],
                                  lifetime=line[store.lifetime.value],
                                  maintenance=line[store.maintenance.value],
                                  installation=line[store.installation.value],
                                  planification=line[store.planification.value],
                                  invest_base=line[store.invest_base.value],
                                  invest_cap=line[store.invest_cap.value],
                                  heat_impact=line[store.heat_impact.value],
                                  elec_impact=line[store.elec_impact.value],
                                  impact_cap=line[store.impact_cap.value]
                                  )
            )
        return list_of_demands


@_dc.dataclass()
class PVConverter(ScenarioToVisualizerAbstract):
    building: int
    peripheral_losses: float
    latitude: float
    longitude: float
    tilt: float
    azimuth: float
    delta_temp_n: int
    capacity_max: float
    capacity_min: float
    lifetime: float
    maintenance: float
    installation: float
    planification: float
    invest_base: float
    invest_cap: float
    heat_impact: float
    elec_impact: float
    impact_cap: float
    roof_area: float
    zenith_angle: float

    def __post_init__(self):
        super().__post_init__()
        if self.from_node:
            raise Warning(f'PV do not have a "from" node. Received {self.from_node}.')

    def get_nodal_infos(self) -> _tp.Optional[dict[str, dict[str, _tp.Union[str, int, float, _pl.Path]]]]:
        if self.active:
            return {"data": {'id': self.id, 'label': self.label, solar.building.value: self.building,
                             solar.peripheral_losses.value: self.peripheral_losses,
                             solar.latitude.value: self.latitude,
                             solar.longitude.value: self.longitude,
                             solar.tilt.value: self.tilt,
                             solar.azimuth.value: self.azimuth,
                             solar.delta_temp_n.value: self.delta_temp_n,
                             solar.capacity_max.value: self.capacity_max,
                             solar.capacity_min.value: self.capacity_min,
                             solar.lifetime.value: self.lifetime,
                             solar.maintenance.value: self.maintenance,
                             solar.installation.value: self.installation,
                             solar.planification.value: self.planification,
                             solar.invest_base.value: self.invest_base,
                             solar.invest_cap.value: self.invest_cap,
                             solar.heat_impact.value: self.heat_impact,
                             solar.elec_impact.value: self.elec_impact,
                             solar.impact_cap.value: self.impact_cap,
                             solar.roof_area.value: self.roof_area,
                             solar.zenith_angle.value: self.zenith_angle,
                             'color': self.color,
                             },
                    "classes": "solar"}


@_dc.dataclass()
class SolarCollectorConverter(ScenarioToVisualizerAbstract):
    building: int
    connect: _tp.Union[str, _abc.Sequence[str]]
    electrical_consumption: float
    peripheral_losses: float
    latitude: float
    longitude: float
    tilt: float
    azimuth: float
    eta_0: float
    a_1: float
    a_2: float
    temp_collector_inlet: float
    delta_temp_n: float
    capacity_max: float
    capacity_min: float
    lifetime: float
    maintenance: float
    installation: float
    planification: float
    invest_base: float
    invest_cap: float
    heat_impact: float
    elec_impact: float
    impact_cap: float
    roof_area: float
    zenith_angle: float

    def __post_init__(self):
        super().__post_init__()

    def get_nodal_infos(self):
        if self.active:
            return {"data": {'id': self.id, 'label': self.label, solar.building.value: self.building,
                             solar.electrical_consumption.value: self.electrical_consumption,
                             solar.peripheral_losses.value: self.peripheral_losses,
                             solar.latitude.value: self.latitude,
                             solar.longitude.value: self.longitude,
                             solar.tilt.value: self.tilt,
                             solar.azimuth.value: self.azimuth,
                             solar.eta_0.value: self.eta_0,
                             solar.a_1.value: self.a_1,
                             solar.a_2.value: self.a_2,
                             solar.temp_collector_inlet.value: self.temp_collector_inlet,
                             solar.delta_temp_n.value: self.delta_temp_n,
                             solar.capacity_max.value: self.capacity_max,
                             solar.capacity_min.value: self.capacity_min,
                             solar.lifetime.value: self.lifetime,
                             solar.maintenance.value: self.maintenance,
                             solar.installation.value: self.installation,
                             solar.planification.value: self.planification,
                             solar.invest_base.value: self.invest_base,
                             solar.invest_cap.value: self.invest_cap,
                             solar.heat_impact.value: self.heat_impact,
                             solar.elec_impact.value: self.elec_impact,
                             solar.impact_cap.value: self.impact_cap,
                             solar.roof_area.value: self.roof_area,
                             solar.zenith_angle.value: self.zenith_angle,
                             'color': self.color,
                             },
                    "classes": "solar"}

    def get_edge_infos(self) -> list[dict[str, dict[str, _tp.Union[str, float, int]]]]:
        if not self.active:
            return []

        all_normal_edges = super().get_edge_infos()

        edges_from_connect_column = []
        if self.connect:
            if not isinstance(self.connect, list):
                self.connect = [c for c in self.connect.split(",")]

            for connect_from in self.connect:
                connect_from = ScenarioToVisualizerAbstract.get_id_with_building(connect_from, self.building)
                energy_type = get_energy_type_based_on_both_labels(connect_from, self.id)
                edges_from_connect_column.append({'data': {'source': connect_from, 'target': self.id},
                                                  "classes": energy_type.value})

        return all_normal_edges + edges_from_connect_column


class SolarConverter(ScenarioToVisualizerAbstract):
    """ This class is only used to prepare the actual PV and SolarCollector convertors.
        As such, only set_from_dataFrame should be used.
    """

    @staticmethod
    def set_from_dataFrame(df: _pd.DataFrame, nr_of_buildings: int):
        list_of_solar = []

        ScenarioToVisualizerAbstract.replace_nans(df)

        if 'active' not in df.columns:
            df['active'] = True

        if 'roof_area' not in df.columns:
            df['roof_area'] = False

        if 'zenith_angle' not in df.columns:
            df['zenith_angle'] = False

        for i, line in df.iterrows():
            if not line['active']:
                continue

            from_nodes, to_nodes = ScenarioToVisualizerAbstract.get_to_and_from_nodes(line)

            label = line['label']
            if label.startswith(SolarTypes.pv):
                list_of_solar.append(
                    SolarConverter.get_PVConverter(line, to_nodes)
                )
            elif label.startswith(SolarTypes.solarCollector):
                list_of_solar.append(
                    SolarConverter.get_SolarCollectorConverter(line, from_nodes, to_nodes)
                )
            else:
                raise ValueError(f'Received unknown solar technology type: {label}')

        return list_of_solar

    @staticmethod
    def get_PVConverter(line, to_nodes):
        return PVConverter(line['label'], None, to_nodes,
                           building=line[solar.building.value], active=True,
                           peripheral_losses=line[solar.peripheral_losses.value],
                           latitude=line[solar.latitude.value],
                           longitude=line[solar.longitude.value],
                           tilt=line[solar.tilt.value],
                           azimuth=line[solar.azimuth.value],
                           delta_temp_n=line[solar.delta_temp_n.value],
                           capacity_max=line[solar.capacity_max.value],
                           capacity_min=line[solar.capacity_min.value],
                           lifetime=line[solar.lifetime.value],
                           maintenance=line[solar.maintenance.value],
                           installation=line[solar.installation.value],
                           planification=line[solar.planification.value],
                           invest_base=line[solar.invest_base.value],
                           invest_cap=line[solar.invest_cap.value],
                           heat_impact=line[solar.heat_impact.value],
                           elec_impact=line[solar.elec_impact.value],
                           impact_cap=line[solar.impact_cap.value],
                           roof_area=line[solar.roof_area.value],
                           zenith_angle=line[solar.zenith_angle.value]
                           )

    @staticmethod
    def get_SolarCollectorConverter(line, from_nodes, to_nodes):
        return SolarCollectorConverter(line['label'], from_nodes, to_nodes,
                                       building=line[solar.building.value], active=True,
                                       connect=line[solar.connect.value],
                                       peripheral_losses=line[solar.peripheral_losses.value],
                                       latitude=line[solar.latitude.value],
                                       longitude=line[solar.longitude.value],
                                       tilt=line[solar.tilt.value],
                                       azimuth=line[solar.azimuth.value],
                                       delta_temp_n=line[solar.delta_temp_n.value],
                                       capacity_max=line[solar.capacity_max.value],
                                       capacity_min=line[solar.capacity_min.value],
                                       lifetime=line[solar.lifetime.value],
                                       maintenance=line[solar.maintenance.value],
                                       installation=line[solar.installation.value],
                                       planification=line[solar.planification.value],
                                       invest_base=line[solar.invest_base.value],
                                       invest_cap=line[solar.invest_cap.value],
                                       heat_impact=line[solar.heat_impact.value],
                                       elec_impact=line[solar.elec_impact.value],
                                       impact_cap=line[solar.impact_cap.value],
                                       electrical_consumption=line[solar.electrical_consumption.value],
                                       eta_0=line[solar.eta_0.value],
                                       a_1=line[solar.a_1.value],
                                       a_2=line[solar.a_2.value],
                                       temp_collector_inlet=line[solar.temp_collector_inlet.value],
                                       roof_area=line[solar.roof_area.value],
                                       zenith_angle=line[solar.zenith_angle.value]
                                       )


@_dc.dataclass()
class LinksConverter(ScenarioToVisualizerAbstract):
    efficiency: float
    invest_base: float
    invest_cap: float
    investment: float
    nr_of_buildings: int

    def __post_init__(self):
        super().__post_init__()

    def get_nodal_infos(self):
        if self.active:
            return {'data': {'id': self.id, 'label': self.label,
                             link.efficiency.value: self.efficiency,
                             link.invest_base.value: self.invest_base,
                             link.invest_cap.value: self.invest_cap,
                             link.investment.value: self.investment,
                             },
                    'classes': 'link'}

    @staticmethod
    def get_to_and_from_nodes(line, nr_of_buildings: int):
        label = line['label']

        to_nodes = []
        from_nodes = []

        for building in range(1, nr_of_buildings + 1):
            if "sh" in label:
                from_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.spaceHeatingBus, building)
                )
                to_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.shDemandBus, building)
                )
            elif "dhw" in label:
                from_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.domesticHotWaterBus, building)
                )
                to_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.dhwDemandBus, building)
                )
            elif "lowTempHeat" in label:
                from_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.lowTSourceBus, building)
                )
                to_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.lowTSinkBus, building)
                )

            elif "heat" in label:
                from_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.heatBus+label[-1], building)
                )
                to_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.heatDemandBus+label[-1], building)
                )
            elif "dhLink" in label:
                from_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.dhHeatInBus, building)
                )
                to_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.dhHeatBus, building)
                )
            else:  # electricity
                from_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.electricityBus, building)
                )
                to_nodes.append(
                    ScenarioToVisualizerAbstract.get_id_with_building(BusTypes.electricityInBus, building)
                )

        return from_nodes, to_nodes

    @staticmethod
    def set_from_dataFrame(df: _pd.DataFrame, nr_of_buildings: int):
        list_of_links = []

        ScenarioToVisualizerAbstract.replace_nans(df)

        if 'active' not in df.columns:
            df['active'] = True

        if link.efficiency.value not in df.columns:
            df[link.efficiency.value] = None

        for i, line in df.iterrows():
            if not line['active']:
                continue

            to_nodes, from_nodes = LinksConverter.get_to_and_from_nodes(line, nr_of_buildings)

            list_of_links.append(
                LinksConverter(
                    line['label'], to_nodes, from_nodes,
                    active=line['active'],
                    efficiency=line[link.efficiency.value],
                    invest_base=line[link.invest_base.value],
                    invest_cap=line[link.invest_cap.value],
                    investment=line[link.investment.value],
                    nr_of_buildings=nr_of_buildings
                )
            )

        return list_of_links



import pandas as pd

import optihood.IO.groupScenarioWriter as _gsw
import optihood.IO.readers as _rd


def create_scenario_file(configFilePath, building, numberOfBuildings=1):
    """
    function to create the input excel file from a config file
    saves the generated excel file at the path given by excel_file_path
    """

    configData = _rd.parse_config(configFilePath)

    excel_data = {}
    buses, columnNames, sheetToSection, updatedLabels = _gsw.get_excel_variable_names()

    buses['domestichotwaterdemand'] = ['', 'domesticHotWaterBus', '']

    temp_h = {}  # to store temp_h values for stratified storage parameters sheet
    FeedinTariff = 0
    profiles = pd.DataFrame(columns=['name', 'path'])

    for sheet in columnNames:
        excel_data[sheet] = pd.DataFrame(columns=columnNames[sheet])
        if sheet == 'stratified_storage':
            newRow = pd.DataFrame(columns=columnNames[sheet])
            newRow.at[0, 'label'] = list(temp_h)[0]
            newRow.at[0, 'temp_h'] = list(temp_h.values())[0]
            newRow.at[1, 'label'] = list(temp_h)[1]
            newRow.at[1, 'temp_h'] = list(temp_h.values())[1]
        for item in configData[sheetToSection[sheet].lower()]:
            type = item[0]
            if item[1] in ['True', 'False']:  # defines whether or not a component is to be added
                if item[1] == 'True':
                    newRow = pd.DataFrame(columns=columnNames[sheet])
                    newRow.at[0, 'label'] = updatedLabels[item[0]]
                    newRow.at[0, 'active'] = int(bool(item[1] == 'True'))
                    # sheet specific options
                    cols = newRow.columns
                    if 'building' in cols:
                        newRow.at[0, 'building'] = 1
                    if 'to' in cols:
                        newRow.at[0, 'to'] = buses[type][0]
                    if 'from' in cols:
                        newRow.at[0, 'from'] = buses[type][1]
                    if 'connect' in cols:
                        newRow.at[0, 'connect'] = buses[type][2]
                    params = configData[type]
                    if sheet == 'commodity_sources':
                        index = [x for x, y in enumerate(params) if y[0] == 'cost'][0]
                        newRow.at[0, 'variable costs'] = params[index][1]
                        index = [x for x, y in enumerate(params) if y[0] == 'impact'][0]
                        newRow.at[0, 'CO2 impact'] = params[index][1]
                        if type == 'electricityresource':
                            index = [x for x, y in enumerate(params) if y[0] == 'feedintariff'][0]
                            FeedinTariff = -float(params[index][1])
                    else:
                        for p in params:
                            if sheet == 'transformers' and p[0] == 'capacity_max':
                                newRow.at[0, 'capacity_DHW'] = p[1]
                                newRow.at[0, 'capacity_SH'] = p[1]
                                if updatedLabels[item[0]] == 'CHP':
                                    newRow.at[0, 'capacity_el'] = str(round(
                                        float(p[1]) * float(newRow.at[0, 'efficiency'].split(',')[0]) / float(
                                            newRow.at[0, 'efficiency'].split(',')[1]), 0))
                            elif sheet == 'storages' and p[0] == 'temp_h':
                                temp_h[newRow.at[0, 'label']] = p[1]
                            else:
                                newRow.at[0, p[0]] = p[1]
                    excel_data[sheet] = pd.concat([excel_data[sheet], newRow])
            elif type == 'path':  # demand path
                buildingFolder = [i[1] for i in configData['demands'] if i[0] == 'folders'][0].split(',')[building]
                buildingPath = item[1] + '\\' + buildingFolder
                profiles = pd.concat([profiles, pd.DataFrame({'name': [updatedLabels[type]], 'path': [buildingPath]})])
            elif 'path' in type:  # weather path
                profiles = pd.concat([profiles, pd.DataFrame({'name': [updatedLabels[type]], 'path': [item[1]]})])
            elif sheet == 'demand' and type != 'folders':
                newRow = pd.DataFrame(columns=columnNames[sheet])
                for d in ['electricityDemand', 'spaceHeatingDemand', 'domesticHotWaterDemand']:
                    newRow.at[0, 'label'] = d
                    newRow.at[0, 'from'] = buses[d.lower()][1]
                    if d == 'spaceHeatingDemand' and item[1] == 0:
                        newRow.at[0, 'fixed'] = item[1]
                        newRow.at[0, 'building model'] = 'Yes'
                    else:
                        newRow.at[0, 'fixed'] = 1
                    excel_data[sheet] = pd.concat([excel_data[sheet], newRow])
                excel_data[sheet]['building'] = 1
                excel_data[sheet]['active'] = 1
                excel_data[sheet]['nominal value'] = 1
            elif sheet == 'stratified_storage':
                newRow.at[0, type] = item[1]
                newRow.at[1, type] = item[1]
            else:  # common parameters of all the components of that section
                if type != 'folders':
                    excel_data[sheet][type] = item[1]
        if sheet == 'stratified_storage':
            excel_data[sheet] = pd.concat([excel_data[sheet], newRow])

    excel_data['profiles'] = profiles

    _gsw.add_grid_connections_to_excel_data(excel_data)

    _gsw.add_busses_to_excel_data(FeedinTariff, excel_data)

    _gsw.add_buildings_to_excel_data(excel_data, numberOfBuildings)

    return excel_data



import abc as _abc
import dataclasses as _dc
import pathlib as _pl
import typing as _tp

import pandas as _pd

import optihood.IO.groupScenarioWriter as _gsw
import optihood.IO.individualScenarioWriter as _isw
import optihood.entities as _ent


@_dc.dataclass()
class ScenarioCreator:
    """ Can be used to get data directly without writing to file using 'ScenarioCreator.get_scenario()'.
        The building_nrs are only used for the individual scenarios.
    """
    config_file_path: _pl.Path
    version: _tp.Literal['grouped', 'individual']
    nr_of_buildings: int = 1
    building_nrs: int = 0
    data: dict[str, _pd.DataFrame] = _dc.field(init=False)

    def __post_init__(self):
        if self.version == 'grouped':
            self.get_scenario = self._get_grouped_scenario
        elif self.version == 'individual':
            self.get_scenario = self._get_individual_scenario

    def _get_grouped_scenario(self):
        data = _gsw.create_scenario_file(self.config_file_path, self.nr_of_buildings)
        return data

    def _get_individual_scenario(self):
        data = _isw.create_scenario_file(self.config_file_path, self.building_nrs, self.nr_of_buildings)
        return data

    @_abc.abstractmethod
    def _write_scenario_to_file(self, file_path: _pl.Path) -> None:
        pass

    def write(self, file_path: _pl.Path) -> None:
        self.data = self.get_scenario()
        self._write_scenario_to_file(file_path)


class ScenarioFileWriterExcel(ScenarioCreator):
    def _write_scenario_to_file(self, file_path: _pl.Path) -> None:
        write_prepared_data_and_sheets_to_excel(file_path, self.data)


class ScenarioFileWriterCSV(ScenarioCreator):
    def __post_init__(self):
        super().__post_init__()
        paths = _ent.CsvInputFilePathsRelative
        self.relative_file_paths = {
            _ent.NodeKeys.buses: paths.buses,
            _ent.NodeKeys.grid_connection: paths.grid_connection,
            _ent.NodeKeys.commodity_sources: paths.commodity_sources,
            _ent.NodeKeys.solar: paths.solar,
            _ent.NodeKeys.transformers: paths.transformers,
            _ent.NodeKeys.demand: paths.demand,
            _ent.NodeKeys.storages: paths.storages,
            _ent.NodeKeys.stratified_storage: paths.stratified_storage,
            _ent.NodeKeys.profiles: paths.profiles,
            _ent.NodeKeys.links: paths.links,
        }

    def _write_scenario_to_file(self, folder_path: _pl.Path) -> None:
        for key, sheet in self.data.items():
            file_path = folder_path / self.relative_file_paths[key]
            write_to_csv(file_path, sheet)


def write_prepared_data_and_sheets_to_excel(excel_file_path: _pl.Path, excel_data: dict):
    # maybe this can be inlined?
    with _pd.ExcelWriter(excel_file_path, engine='openpyxl') as writer:
        for sheet, data in excel_data.items():
            data.to_excel(writer, sheet_name=sheet, index=False)


def write_to_csv(file_path: _pl.Path, data_sheet: _pd.DataFrame) -> None:
    data_sheet.to_csv(file_path, index=False)



import oemof.solph as solph
import numpy as np
import pandas as pd
import pvlib
import optihood.combined_prod as cp
from oemof.thermal.solar_thermal_collector import flat_plate_precalc
from oemof.solph._plumbing import sequence as solph_sequence
from pyomo.core.base.block import ScalarBlock
from pyomo.environ import BuildAction
from pyomo.environ import Constraint
from optihood._helpers import *

class SolarCollector:
    """
        Currently, this class supports maximum 3 temperature levels
    """
    def __init__(self, label, buildingLabel, inputs, outputs, connectors, electrical_consumption, peripheral_losses, latitude,
                 longitude,
                 collector_tilt, roof_area, zenith_angle, collector_azimuth, eta_0, a_1, a_2, temp_collector_inlet,
                 delta_temp_n, irradiance_global,
                 irradiance_diffuse, temp_amb_col, capacityMin, capacityMax, epc, base, env_capa, env_flow, varc, dispatchMode):

        flatPlateCollectorData_sh = flat_plate_precalc(latitude, longitude, collector_tilt, collector_azimuth, eta_0, a_1, a_2, temp_collector_inlet, delta_temp_n[0],
                irradiance_global, irradiance_diffuse, temp_amb_col)
        self.collectors_heat_sh = flatPlateCollectorData_sh['collectors_heat'] / 1000 # flow in kWh per m² of solar thermal panel
        self.collectors_eta_c_sh = flatPlateCollectorData_sh['eta_c']
        ep_costs = [epc]
        offset = [base]
        if outputs.__len__()>=2:
            flatPlateCollectorData_dhw = flat_plate_precalc(latitude, longitude, collector_tilt, collector_azimuth, eta_0, a_1, a_2, temp_collector_inlet, delta_temp_n[-1],
                irradiance_global, irradiance_diffuse, temp_amb_col)
            self.collectors_heat_dhw = flatPlateCollectorData_dhw['collectors_heat'] / 1000 # flow in kWh per m² of solar thermal panel
            self.collectors_eta_c_dhw = flatPlateCollectorData_dhw['eta_c']
            ep_costs = [0, epc]
            offset = [0.000001, base]
            if outputs.__len__()>2:
                flatPlateCollectorData_T2 = flat_plate_precalc(latitude, longitude, collector_tilt, collector_azimuth, eta_0, a_1, a_2, temp_collector_inlet, delta_temp_n[1],
                irradiance_global, irradiance_diffuse, temp_amb_col)
                self.collectors_heat_T2 = flatPlateCollectorData_T2['collectors_heat'] / 1000 # flow in kWh per m² of solar thermal panel
                self.collectors_eta_c_T2 = flatPlateCollectorData_T2['eta_c']
                ep_costs = [0, 0, epc]
                offset = [0.000001, 0.000001, base]

        if not (np.isnan(roof_area) or np.isnan(zenith_angle)):
            self.surface_used = self._calculateArea(zenith_angle, collector_tilt, collector_azimuth)
        else:
            self.surface_used = np.nan

        if dispatchMode:
            investArgsSH = {'ep_costs': ep_costs[0],
                            'minimum': capacityMin,
                            'maximum': capacityMax,
                            'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used,
                                                      'roof_area': roof_area}}
            if outputs.__len__() >= 2:
                investArgsDHW = {'ep_costs': ep_costs[-1],
                                 'minimum': capacityMin,
                                 'maximum': capacityMax,
                                 'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used,
                                                       'roof_area': roof_area}}
                if outputs.__len__() > 2:
                    investArgsT2 = {'ep_costs': ep_costs[1],
                                    'minimum': capacityMin,
                                    'maximum': capacityMax,
                                    'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used,
                                                          'roof_area': roof_area}}
        else:
            investArgsSH = {'ep_costs': ep_costs[0],
                            'minimum': capacityMin,
                            'maximum': capacityMax,
                            'nonconvex': True,
                            'offset': offset[0],
                            'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used,
                                                  'roof_area': roof_area}}
            if outputs.__len__() >= 2:
                investArgsDHW = {'ep_costs': ep_costs[-1],
                                 'minimum': capacityMin,
                                 'maximum': capacityMax,
                                 'nonconvex': True,
                                 'offset': offset[-1],
                                 'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used,
                                                       'roof_area': roof_area}}
                if outputs.__len__() > 2:
                    investArgsT2 = {'ep_costs': ep_costs[1],
                                    'minimum': capacityMin,
                                    'maximum': capacityMax,
                                    'nonconvex': True,
                                    'offset': offset[1],
                                    'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used,
                                                          'roof_area': roof_area}}

        self.__heat_source_sh = solph.components.Source(
            label='heatSource_SH' + label + "__" + buildingLabel,
            outputs={
                connectors[0]: solph.Flow(
                    fix=self.collectors_heat_sh,
                    investment=solph.Investment(**investArgsSH),
                    variable_costs=varc,
                    custom_attributes={'env_per_flow': env_flow},
                )
            },
        )
        self.__excessheat_sh = solph.components.Sink(
            label='excessheat_SH' + label + "__" + buildingLabel, inputs={connectors[0]: solph.Flow()}
        )
        self.__heat_transformer_sh = solph.components.Transformer(
            label=label + 'SH__' + buildingLabel,
            inputs={connectors[0]: solph.Flow(), inputs: solph.Flow()},
            outputs={outputs[0]: solph.Flow()},
            conversion_factors={
                connectors[0]: 1,
                inputs: electrical_consumption * (1 - peripheral_losses),
                outputs[0]: 1 - peripheral_losses
            },
        )

        if outputs.__len__() >= 2:
            self.__heat_source_dhw = solph.components.Source(
                label='heatSource_DHW' + label + "__" + buildingLabel,
                outputs={
                    connectors[-1]: solph.Flow(
                        fix=self.collectors_heat_dhw,
                        investment=solph.Investment(**investArgsDHW),
                        variable_costs=varc,
                        custom_attributes={'env_per_flow': env_flow},
                    )
                },
            )
            self.__excessheat_dhw = solph.components.Sink(
                label='excessheat_DHW' + label + "__" + buildingLabel, inputs={connectors[-1]: solph.Flow()}
            )
            self.__heat_transformer_dhw = solph.components.Transformer(
                label=label + 'DHW__' + buildingLabel,
                inputs={connectors[-1]: solph.Flow(), inputs: solph.Flow()},
                outputs={outputs[-1]: solph.Flow()},
                conversion_factors={
                    connectors[-1]: 1,
                    inputs: electrical_consumption * (1 - peripheral_losses),
                    outputs[-1]: 1 - peripheral_losses
                },
            )
            if outputs.__len__()>2:
                self.__heat_source_T2 = solph.components.Source(
                    label='heatSource_T2' + label + "__" + buildingLabel,
                    outputs={
                        connectors[1]: solph.Flow(
                            fix=self.collectors_heat_dhw,
                            investment=solph.Investment(**investArgsT2),
                            variable_costs=varc,
                            custom_attributes={'env_per_flow': env_flow},
                        )
                    },
                )
                self.__excessheat_T2 = solph.components.Sink(
                    label='excessheat_T2' + label + "__" + buildingLabel, inputs={connectors[1]: solph.Flow()}
                )
                self.__heat_transformer_T2 = solph.components.Transformer(
                    label=label + 'T2__' + buildingLabel,
                    inputs={connectors[1]: solph.Flow(), inputs: solph.Flow()},
                    outputs={outputs[1]: solph.Flow()},
                    conversion_factors={
                        connectors[1]: 1,
                        inputs: electrical_consumption * (1 - peripheral_losses),
                        outputs[1]: 1 - peripheral_losses
                    },
                )
            else:
                self.__heat_source_T2 = None
                self.__excessheat_T2 = None
                self.__heat_transformer_T2 = None
        else:
            self.__heat_source_dhw = None
            self.__heat_transformer_dhw = None
            self.__excessheat_dhw = None
            self.__heat_source_T2 = None
            self.__excessheat_T2 = None
            self.__heat_transformer_T2 = None

    def getSolar(self, type):
        if type == 'source':
            return self.__heat_source_sh, self.__heat_source_T2, self.__heat_source_dhw
        elif type == 'transformer':
            return self.__heat_transformer_sh, self.__heat_transformer_T2, self.__heat_transformer_dhw
        elif type == 'sink':
            return self.__excessheat_sh, self.__excessheat_T2, self.__excessheat_dhw
        else:
            print("Label not identified...")
            return []

    def _calculateArea(self, zenith_angle, collector_tilt, collector_azimuth):
        # Coefficient representing the area used by one unit of capacity of a solar collector
        coeff = -np.sin((zenith_angle+collector_tilt)*np.pi/180)*np.cos(collector_azimuth*np.pi/180)/np.sin(zenith_angle*np.pi/180)
        return coeff

class PVT:
    """
    Currently, this class supports maximum 3 temperature levels
    """
    def __init__(self, label, buildingLabel, inputs, outputs, connectors, electrical_consumption, peripheral_losses, latitude,
                 longitude,
                 collector_tilt, roof_area, zenith_angle, collector_azimuth, eta_0, a_1, a_2, temp_collector_inlet,temp_amb,
                 delta_temp_n, irradiance_global,
                 irradiance_diffuse, capacityMin, capacityMax, epc, base, env_capa, env_flow, varc, dispatchMode,
                 pv_efficiency=0.15, taualpha=0.85):

        pvdata = self.computePvSolarPosition(irradiance_diffuse, irradiance_global, latitude, longitude, collector_azimuth,
                                           collector_tilt, temp_amb)
        if outputs.__len__()>=3:
            temp_pv = delta_temp_n[-1]
            outputPV = outputs[-1]
        else:
            temp_pv = delta_temp_n[0]
            outputPV = outputs[1]
        pv_electricity = np.minimum(self.pvtElPrecalc(temp_amb, temp_collector_inlet, pvdata['pv_ira'] / 1000, temp_pv), capacityMax)
        pvtCollectorData_sh = self.pvtThPrecalc(latitude, longitude, collector_tilt, collector_azimuth, eta_0, a_1, a_2,
                                                temp_collector_inlet, delta_temp_n[0],
                                       irradiance_global, irradiance_diffuse, temp_amb,  pv_efficiency, taualpha)
        self.collectors_heat_sh = pvtCollectorData_sh['collectors_heat'] / 1000
        self.collectors_eta_c_sh = pvtCollectorData_sh['eta_c']
        ep_costs = [epc]
        offset = [base]
        if outputs.__len__()>=3:
            pvtCollectorData_dhw = self.pvtThPrecalc(latitude, longitude, collector_tilt, collector_azimuth, eta_0, a_1, a_2,
                                                 temp_collector_inlet, delta_temp_n[-1],
                                                 irradiance_global, irradiance_diffuse, temp_amb, pv_efficiency, taualpha)
            self.collectors_heat_dhw = pvtCollectorData_dhw['collectors_heat'] / 1000
            self.collectors_eta_c_dhw = pvtCollectorData_dhw['eta_c']
            ep_costs = [0, epc]
            offset = [0.000001, base]
            if outputs.__len__()>3:
                pvtCollectorData_T2 = self.pvtThPrecalc(latitude, longitude, collector_tilt, collector_azimuth, eta_0,
                                                         a_1, a_2,
                                                         temp_collector_inlet, delta_temp_n[1],
                                                         irradiance_global, irradiance_diffuse, temp_amb, pv_efficiency,
                                                         taualpha)
                self.collectors_heat_T2 = pvtCollectorData_T2['collectors_heat'] / 1000
                self.collectors_eta_c_T2 = pvtCollectorData_T2['eta_c']
                ep_costs = [0, 0, epc]
                offset = [0.000001, 0.000001, base]

        self.taualpha = taualpha


        self.surface_used = self.calculateArea(zenith_angle, collector_tilt, collector_azimuth)
        surface_used_el = self.calculateArea(zenith_angle, collector_tilt, collector_azimuth, pv_efficiency)
        if dispatchMode:
            investArgsEl = {'ep_costs':0,
                         'minimum':capacityMin,
                         'maximum':capacityMax,                     
                         'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used, 'space_el': surface_used_el,
                                                  'roof_area': roof_area}}
            investArgsSH = {'ep_costs': ep_costs[0],
                          'minimum': capacityMin,
                          'maximum': capacityMax,
                          'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used, 'space_el': surface_used_el,
                                                  'roof_area': roof_area}}
            if outputs.__len__() >= 3:
                investArgsDHW = {'ep_costs': ep_costs[-1],
                              'minimum': capacityMin,
                              'maximum': capacityMax,
                              'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used, 'space_el': surface_used_el,
                                                      'roof_area': roof_area}}
                if outputs.__len__() > 3:
                    investArgsT2 = {'ep_costs': ep_costs[1],
                                  'minimum': capacityMin,
                                  'maximum': capacityMax,
                                  'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used, 'space_el': surface_used_el,
                                                          'roof_area': roof_area}}

        else:
            investArgsEl = {'ep_costs': 0,
                            'minimum': capacityMin,
                            'maximum': capacityMax,
                            'nonconvex': True,
                            'offset': 0.000001,
                            'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used, 'space_el': surface_used_el,
                                                  'roof_area': roof_area}}
            investArgsSH = {'ep_costs': ep_costs[0],
                            'minimum': capacityMin,
                            'maximum': capacityMax,
                            'nonconvex': True,
                            'offset': offset[0],
                            'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used, 'space_el': surface_used_el,
                                                  'roof_area': roof_area}}
            if outputs.__len__() >= 3:
                investArgsDHW = {'ep_costs': ep_costs[-1],
                                 'minimum': capacityMin,
                                 'maximum': capacityMax,
                                 'nonconvex': True,
                                 'offset': offset[-1],
                                 'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used, 'space_el': surface_used_el,
                                                      'roof_area': roof_area}}
                if outputs.__len__() > 3:
                    investArgsT2 = {'ep_costs': ep_costs[1],
                            'minimum': capacityMin,
                            'maximum': capacityMax,
                            'nonconvex': True,
                            'offset': offset[1],
                            'custom_attributes': {'env_per_capa': env_capa, 'space': self.surface_used, 'space_el': surface_used_el,
                                                  'roof_area': roof_area}}


        self.__PVTel_source = solph.components.Source(label='elSource_' + label + '__' + buildingLabel,
                                 outputs={outputPV: solph.Flow(
                                     investment=solph.Investment(**investArgsEl),
                                     variable_costs=varc,
                                     max=pv_electricity,
                                     custom_attributes={'env_per_flow': env_flow}
                                 )}
                                 )
        self.__PVTheat_source_sh = solph.components.Source(
            label='heatSource_SH' + label + "__" + buildingLabel,
            outputs={
                connectors[0]: solph.Flow(
                    fix=self.collectors_heat_sh,
                    investment=solph.Investment(**investArgsSH),
                    variable_costs=varc,
                    custom_attributes={'env_per_flow': env_flow},
                )
            },
        )
        self.__PVT_excessheat_sh = solph.components.Sink(
            label='excessheat_SH' + label + "__" + buildingLabel, inputs={connectors[0]: solph.Flow()}
        )
        self.__PVTheat_transformer_sh = solph.components.Transformer(
            label=label + 'SH__' + buildingLabel,
            inputs={connectors[0]: solph.Flow(), inputs: solph.Flow()},
            outputs={outputs[0]: solph.Flow()},
            conversion_factors={
                connectors[0]: 1,
                inputs: electrical_consumption * (1 - peripheral_losses),
                outputs[0]: 1 - peripheral_losses
            },
        )

        if outputs.__len__() >= 3:
            self.__PVTheat_source_dhw = solph.components.Source(
                label='heatSource_DHW' + label + "__" + buildingLabel,
                outputs={
                    connectors[-1]: solph.Flow(
                        fix=self.collectors_heat_dhw,
                        investment=solph.Investment(**investArgsDHW),
                        variable_costs=varc,
                        custom_attributes={'env_per_flow': env_flow},
                    )
                },
            )
            self.__PVT_excessheat_dhw = solph.components.Sink(
                label='excessheat_DHW' + label + "__" + buildingLabel, inputs={connectors[-1]: solph.Flow()}
            )
            self.__PVTheat_transformer_dhw = solph.components.Transformer(
                label=label + 'DHW__' + buildingLabel,
                inputs={connectors[-1]: solph.Flow(), inputs: solph.Flow()},
                outputs={outputs[-2]: solph.Flow()},
                conversion_factors={
                    connectors[-1]: 1,
                    inputs: electrical_consumption * (1 - peripheral_losses),
                    outputs[-2]: 1 - peripheral_losses
                },
            )
            if outputs.__len__()>3:
                self.__PVTheat_source_T2 = solph.components.Source(
                    label='heatSource_T2' + label + "__" + buildingLabel,
                    outputs={
                        connectors[1]: solph.Flow(
                            fix=self.collectors_heat_dhw,
                            investment=solph.Investment(**investArgsT2),
                            variable_costs=varc,
                            custom_attributes={'env_per_flow': env_flow},
                        )
                    },
                )
                self.__PVT_excessheat_T2 = solph.components.Sink(
                    label='excessheat_T2' + label + "__" + buildingLabel, inputs={connectors[1]: solph.Flow()}
                )
                self.__PVTheat_transformer_T2 = solph.components.Transformer(
                    label=label + 'T2__' + buildingLabel,
                    inputs={connectors[1]: solph.Flow(), inputs: solph.Flow()},
                    outputs={outputs[1]: solph.Flow()},
                    conversion_factors={
                        connectors[1]: 1,
                        inputs: electrical_consumption * (1 - peripheral_losses),
                        outputs[1]: 1 - peripheral_losses
                    },
                )
            else:
                self.__PVTheat_source_T2 = None
                self.__PVT_excessheat_T2 = None
                self.__PVTheat_transformer_T2 = None
        else:
            self.__PVTheat_source_dhw = None
            self.__PVTheat_transformer_dhw = None
            self.__PVT_excessheat_dhw = None
            self.__PVTheat_source_T2 = None
            self.__PVT_excessheat_T2 = None
            self.__PVTheat_transformer_T2 = None

    def pvtThPrecalc(self, latitude, longitude, collector_tilt, collector_azimuth, eta_0, a_1, a_2, temp_collector_inlet, delta_temp_n,
            irradiance_global, irradiance_diffuse, temp_amb,  pv_efficiency, taualpha):
        # function inspired from oemof thermal solar_thermal_collector
        # The calculation of eta_c is overriden
        data = pd.DataFrame(
            {
                'ghi': irradiance_global,
                'dhi': irradiance_diffuse,
                'temp_amb': temp_amb
            }
        )
        solposition = pvlib.solarposition.get_solarposition(
            time=data.index, latitude=latitude, longitude=longitude
        )
        dni = pvlib.irradiance.dni(
            ghi=data['ghi'], dhi=data['dhi'], zenith=solposition['apparent_zenith']
        )
        total_irradiation = pvlib.irradiance.get_total_irradiance(
            surface_tilt=collector_tilt,
            surface_azimuth=collector_azimuth,
            solar_zenith=solposition['apparent_zenith'],
            solar_azimuth=solposition['azimuth'],
            dni=dni.fillna(0),  # fill NaN values with '0'
            ghi=data['ghi'],
            dhi=data['dhi'],
        )
        data['col_ira'] = total_irradiation['poa_global']
        data["eta_c"] = self.calcEtaC(eta_0, a_1, a_2, temp_collector_inlet, delta_temp_n, data['temp_amb'], total_irradiation['poa_global'], pv_efficiency, taualpha)
        collectors_heat = data["eta_c"] * total_irradiation['poa_global']
        data["collectors_heat"] = collectors_heat
        return data

    def calcEtaC(self, eta_0, a_1, a_2, temp_collector_inlet, delta_temp_n, temp_amb, irradiation, pv_efficiency, taualpha):
        delta_t = temp_collector_inlet + delta_temp_n - temp_amb
        eta_c = pd.Series()
        for index, value in irradiation.items():
            if value > 0:
                eta = (eta_0*(1-pv_efficiency/(taualpha))
                        - a_1 * delta_t[index] / value
                        - a_2 * delta_t[index] ** 2 / value)
                eta_c[index] = eta*(eta > 0) + 0
            else:
                eta_c[index] = 0
        return eta_c

    def computePvSolarPosition(self, irradiance_diffuse, irradiance_global, latitude, longitude, pv_azimuth, pv_tilt,
                               temp_amb_pv):
        data = pd.DataFrame(
            {
                'ghi': irradiance_global,
                'dhi': irradiance_diffuse,
                'temp_amb': temp_amb_pv
            }
        )
        solposition = pvlib.solarposition.get_solarposition(
            time=data.index, latitude=latitude, longitude=longitude
        )
        dni = pvlib.irradiance.dni(
            ghi=data['ghi'], dhi=data['dhi'], zenith=solposition['apparent_zenith']
        )
        total_irradiation = pvlib.irradiance.get_total_irradiance(
            surface_tilt=pv_tilt,
            surface_azimuth=pv_azimuth,
            solar_zenith=solposition['apparent_zenith'],
            solar_azimuth=solposition['azimuth'],
            dni=dni.fillna(0),  # fill NaN values with '0'
            ghi=data['ghi'],
            dhi=data['dhi'],
        )
        data['pv_ira'] = total_irradiation['poa_global']
        return data

    # model according to TriHP Energy management algorithms description. D6.2 v1.0
    # nominal power is 1 KW due to the normed optimizer
    def pvtElPrecalc(self, temp_amb, temp_collector_inlet, i_H_t,delta_temp_n, a4=-0.062059,
                   a5=0.04277774, a6=9.692792, a7=-1.885868, a8=6.6):
        temp_cell = temp_collector_inlet + delta_temp_n - temp_amb
        pvPower = np.maximum(0, (a4*i_H_t + a5)*temp_cell + a6*i_H_t + a7) / a8
        return pvPower


    def calculateArea(self, zenith_angle, pv_tilt, pv_azimuth, pv_efficiency=0.0):
        # Coefficient representing the area used by one unit of capacity (kW thermal) of a PVT panel
        coeff = -np.sin((zenith_angle+pv_tilt)*np.pi/180)*np.cos(pv_azimuth*np.pi/180)/np.sin(zenith_angle*np.pi/180)
        if pv_efficiency:
            coeff /= pv_efficiency
        return coeff

    def getPVT(self, type):
        if type == 'heat_source':
            return self.__PVTheat_source_sh, self.__PVTheat_source_T2, self.__PVTheat_source_dhw
        elif type == 'heat_transformer':
            return self.__PVTheat_transformer_sh, self.__PVTheat_transformer_T2, self.__PVTheat_transformer_dhw
        elif type == 'el_source':
            return self.__PVTel_source
        elif type == 'excess_heat_sink':
            return self.__PVT_excessheat_sh, self.__PVT_excessheat_T2, self.__PVT_excessheat_dhw
        else:
            print("Label not identified...")
            return []




class HeatPumpLinear:
    """
    Generic linear heat pump model.
    Allows the user to input the necessary coefficients for COP calculation.
    """
    def __init__(self, label, operationTemperatures, temperatureLow, coef_W, coef_Q, input, output,
                 capacityMin, capacityMax, nomEff,
                 epc, base, varc, env_flow, env_capa, dispatchMode):
        outputTemperatures = {}
        for i in range(len(output)):
            outputTemperatures[output[i]] = operationTemperatures[i]
        self.cop = {o:self._calculate_cop(t, temperatureLow, coef_W, coef_Q) for o,t in outputTemperatures.items()}
        self.avgCopSh = (sum(self.cop[output[0]])/len(self.cop[output[0]])) # cop at lowest temperature, i.e. temperature of space heating
        self.nominalEff = nomEff
        if dispatchMode:
            investArgs = {'ep_costs' : epc * nomEff,
            'minimum' : capacityMin / nomEff,
            'maximum' : capacityMax / nomEff,
            'custom_attributes': {'env_per_capa': env_capa * nomEff}}
        else:
            investArgs = {'ep_costs' : epc * nomEff,
            'minimum' : capacityMin / nomEff,
            'maximum' : capacityMax / nomEff,
            'nonconvex': True,
            'offset' : base,
            'custom_attributes' : {'env_per_capa': env_capa * nomEff}}
        outputDict = {k: solph.Flow(variable_costs=varc,
                          custom_attributes={'env_per_flow': env_flow}, ) for k in output}
        inputDict = {input[0]: solph.Flow(investment=solph.Investment(**investArgs))}
        if len(input) > 1:
            # Two input HP, second input is for Q evaporator
            inputDict.update({input[1]: solph.Flow()})
        self._heatpump = cp.CombinedTransformer(label=label,
                                            inputs=inputDict,
                                            outputs=outputDict,
                                            efficiencies=self.cop)

    @staticmethod
    def _calculate_cop(t_high, t_low, coef_W, coef_Q):
        t_low_K = t_low / 273.15
        t_high_K = t_high / 273.15
        QCondenser = (
                coef_Q[0] +
                coef_Q[1] * t_low_K +
                coef_Q[2] * t_high_K +
                coef_Q[3] * t_low_K * t_high_K +
                coef_Q[4] * (t_low_K ** 2) +
                coef_Q[5] * (t_high_K ** 2)
        )
        WCompressor = (
                coef_W[0] +
                coef_W[1] * t_low_K +
                coef_W[2] * t_high_K +
                coef_W[3] * t_low_K * t_high_K +
                coef_W[4] * (t_low_K ** 2) +
                coef_W[5] * (t_high_K ** 2)
        )
        cop = np.divide(QCondenser, WCompressor)
        return cop

    def getHP(self, type):
        if type == 'sh':
            return self._heatpump
        else:
            print("Transformer label not identified...")
            return []


class Chiller(HeatPumpLinear):
    pass


class CHP:
    "Information about the model can be found in combined_pro.py CombinedCHP"
    def __init__(self, buildingLabel, input, output, efficiency,
                 capacityMin, capacitySH, epc, base, varc1, varc2, env_flow1, env_flow2, env_capa, timesteps, dispatchMode):
        self.avgEff = efficiency[1]
        outputEff = {}
        for i in range(len(output)):
            outputEff[output[i]] = [efficiency[i]] * timesteps
        if dispatchMode:
            investArgs = {'ep_costs': epc * self.avgEff,
                          'minimum': capacityMin / self.avgEff,
                          'maximum': capacitySH / self.avgEff,
                          'custom_attributes': {'env_per_capa': env_capa * self.avgEff}}
        else:
            investArgs={'ep_costs':epc*self.avgEff,
                        'minimum':capacityMin/self.avgEff,
                        'maximum':capacitySH/self.avgEff,
                        'nonconvex':True,
                        'offset':base,
                        'custom_attributes': {'env_per_capa': env_capa * self.avgEff}}
        outputDict = {k: solph.Flow(variable_costs=varc2, custom_attributes={'env_per_flow': env_flow2}) for k in output if "electricity" not in k.label}
        elOutKey = [k for k in output if "electricity" in k.label][0]
        outputDict.update({elOutKey: solph.Flow(variable_costs=varc1, custom_attributes={'env_per_flow': env_flow1})})
        self.__CHP = cp.CombinedCHP(
                        label='CHP'+'__'+buildingLabel,
                        inputs={
                            input: solph.Flow(
                                investment=solph.Investment(**investArgs),
                            )
                        },
                        outputs=outputDict,
                        efficiencies=outputEff
                    )

    def getCHP(self):
        return self.__CHP

class GasBoiler(cp.CombinedTransformer):
    """
    Sunsetted class, will be removed in the next updates..., use class GenericCombinedTransformer instead
    """
    def __init__(self, buildingLabel, input, output, efficiency,
                 capacityMin, capacityMax, epc, base, varc, env_flow, env_capa, dispatchMode):
        self.__efficiency = efficiency[0]
        outputEff = {}
        for i in range(len(output)):
            outputEff[output[i]] = efficiency[i]
        if dispatchMode:
            investArgs= {'ep_costs':epc*efficiency[0],
                        'minimum':capacityMin/efficiency[0],
                        'maximum':capacityMax/efficiency[0],
                        'custom_attributes': {'env_per_capa': env_capa * efficiency[0]}}
        else:
            investArgs= {'ep_costs':epc*efficiency[0],
                        'minimum':capacityMin/efficiency[0],
                        'maximum':capacityMax/efficiency[0],
                        'nonconvex':True,
                        'offset':base,
                        'custom_attributes': {'env_per_capa': env_capa * efficiency[0]}}
        outputDict = {k: solph.Flow(variable_costs=varc, custom_attributes={'env_per_flow': env_flow}) for k in output}
        super(GasBoiler, self).__init__(
            label='GasBoiler'+'__'+buildingLabel,
            inputs={
                input: solph.Flow(
                    investment=solph.Investment(**investArgs),
                )
            },
            outputs=outputDict,
            efficiencies=outputEff
                 )

class GenericCombinedTransformer(cp.CombinedTransformer):
    def __init__(self, label, input, output, efficiency,
                 capacityMin, capacityMax, epc, base, varc, env_flow, env_capa, dispatchMode):
        self.__efficiency = efficiency[0]
        outputEff = {}
        for i in range(len(output)):
            outputEff[output[i]] = efficiency[i]
        if dispatchMode:
            investArgs = {'ep_costs': epc * efficiency[0],
                          'minimum': capacityMin / efficiency[0],
                          'maximum': capacityMax / efficiency[0],
                          'custom_attributes': {'env_per_capa': env_capa * efficiency[0]}}
        else:
            investArgs={'ep_costs':epc*efficiency[0],
                        'minimum':capacityMin/efficiency[0],
                        'maximum':capacityMax/efficiency[0],
                        'nonconvex':True,
                        'offset':base,
                        'custom_attributes': {'env_per_capa': env_capa * efficiency[0]}}
        outputDict = {k: solph.Flow(variable_costs=varc, custom_attributes={'env_per_flow': env_flow}) for k in output}
        super(GenericCombinedTransformer, self).__init__(
            label=label,
            inputs={
                input: solph.Flow(investment=solph.Investment(**investArgs) )
            },
            outputs=outputDict,
            efficiencies=outputEff
                 )


class ElectricRod(cp.CombinedTransformer):
    """
    To be sunsetted in the next versions --> Use class GenericCombinedTransformer instead
    """
    def __init__(self, buildingLabel, input, output, efficiency,
                 capacityMin, capacityMax, epc, base, varc, env_flow, env_capa, dispatchMode):
        self.__efficiency = efficiency
        outputEff = {}
        for i in range(len(output)):
            outputEff[output[i]] = efficiency
        if dispatchMode:
            investArgs = {'ep_costs': epc * efficiency,
                          'minimum': capacityMin / efficiency,
                          'maximum': capacityMax / efficiency,
                          'custom_attributes': {'env_per_capa': env_capa * efficiency}}
        else:
            investArgs={'ep_costs':epc*efficiency,
                        'minimum':capacityMin/efficiency,
                        'maximum':capacityMax/efficiency,
                        'nonconvex':True,
                        'offset':base,
                        'custom_attributes': {'env_per_capa': env_capa * efficiency}}
        outputDict = {k: solph.Flow(variable_costs=varc, custom_attributes={'env_per_flow': env_flow}) for k in output}
        super(ElectricRod, self).__init__(
            label='ElectricRod'+'__'+buildingLabel,
            inputs={
                input: solph.Flow(investment=solph.Investment(**investArgs) )
            },
            outputs=outputDict,
            efficiencies=outputEff
                 )

class GeothermalHeatPumpLinearSingleUse:
    "Class implementing a linear model for geothermal heat pump for single use only (either DHW or SH but not both)"
    def __init__(self, buildingLabel, temperatureH, temperatureLow, input, outputH,
                 capacityMin, capacityMax, epc, base, varc, env_flow, env_capa, dispatchMode):
        self.__copH = self._calculateCop(temperatureH, temperatureLow)
        #self.avgCopSh = (sum(self.__copH)/len(self.__copH))
        #self.nominalEff = nomEff
        if dispatchMode:
            investArgs = {'ep_costs':epc,
                            'minimum':0,
                            'maximum':capacityMax,
                            'custom_attributes': {'env_per_capa': env_capa}}
        else:
            investArgs = {'ep_costs':epc,
                            'minimum':0,
                            'maximum':capacityMax,
                            'nonconvex':True,
                            'offset':base,
                             'custom_attributes': {'env_per_capa': env_capa}}
        self.__geothermalheatpump = solph.components.Transformer(label=f'GWHP{str(temperatureH)}' + '__' + buildingLabel,
                                            inputs={input: solph.Flow()},
                                            outputs={outputH: solph.Flow(
                                                          variable_costs=varc,
                                                          custom_attributes={'env_per_flow': env_flow},
                                                          investment=solph.Investment(**investArgs),
                                                      )
                                                  },
                                            conversion_factors={outputH: self.__copH})

    def _calculateCop(self, tHigh, tLow):
        coefW = [0.1600, -1.2369, 19.9391, 19.3448, 7.1057, -1.4048]
        coefQ = [13.8978, 114.8358, -9.3634, -179.4227, 342.3363, -12.4969]
        QCondenser = coefQ[0] + (coefQ[1] * tLow / 273.15) + (coefQ[2] * tHigh / 273.15) + (
                coefQ[3] * tLow / 273.15 * tHigh / 273.15) + (
                             coefQ[4] * (tLow / 273.15) ** 2) + (
                             coefQ[5] * (tHigh / 273.15) ** 2)
        WCompressor = coefW[0] + (coefW[1] * tLow / 273.15) + (coefW[2] * tHigh / 273.15) + (
                coefW[3] * tLow / 273.15 * tHigh / 273.15) + (
                             coefW[4] * (tLow / 273.15) ** 2) + (
                              coefW[5] * (tHigh / 273.15) ** 2)
        cop = np.divide(QCondenser, WCompressor)
        return cop

    def getHP(self, type):
        if type == 'sh':
            return self.__geothermalheatpump
        else:
            print("Transformer label not identified...")
            return []